from daydream.plugins.kubernetes.nodes.utils import label_key_value_to_node_id

from .base import KubernetesBaseNode


class KubernetesService(KubernetesBaseNode):
    async def get_identifiers(self) -> list[str | None]:
        return [
            *await super().get_identifiers(),
            *self.raw_data.get("spec", {}).get("clusterIPs", []),
        ]

    async def get_reference_identifiers(self) -> list[str | None | tuple[str | None, str]]:
        return [
            *await super().get_reference_identifiers(),
            *[
                (label_key_value_to_node_id(key, value), "has_selector")
                for key, value in self.raw_data.get("spec", {}).get("selector", {}).items()
            ],
            *[
                (i.get("hostname"), "receives_traffic_from")
                for i in self.raw_data.get("status", {}).get("loadBalancer", {}).get("ingress", [])
                if "hostname" in i
            ],
        ]



import random
from collections import defaultdict
from datetime import datetime
from typing import Any

import numpy as np
from drain3.template_miner import TemplateMiner
from drain3.template_miner_config import TemplateMinerConfig
from sparklines import sparklines

from daydream.models import LogAnomaly, LogLine
from daydream.utils import print

MAX_TRAINING_SAMPLE_SIZE = 100_000
MAX_ANALYSIS_SAMPLE_SIZE = 250_000
MAX_ANOMALIES_TO_RETURN = 10


def segment_logs_by_time(
    log_entries: list[LogLine], window_minutes: int = 5
) -> dict[datetime, list[str]]:
    time_windows = defaultdict(list)

    for log_entry in log_entries:
        window_key = log_entry.time.replace(
            minute=log_entry.time.minute - log_entry.time.minute % window_minutes,
            second=0,
            microsecond=0,
        )
        time_windows[window_key].append(log_entry.log)

    return time_windows


def analyze_pattern_trends(
    time_windows: dict[datetime, list[str]], template_miner: TemplateMiner
) -> tuple[dict[str, list[tuple[datetime, float]]], dict[str, list[tuple[datetime, int]]]]:
    """Track pattern frequencies across time windows

    Args:
        time_windows: Dictionary mapping window times to lists of log lines
        template_miner: Pre-trained drain3 TemplateMiner

    Returns:
        pattern_trends: Dictionary mapping patterns to lists of (time, frequency (%)) tuples
        pattern_counts: Dictionary mapping patterns to lists of (time, counts) tuples
    """
    # Get all templates from the pre-processed template miner
    templates_info = {}

    # Extract all templates (cluster info) from the template miner
    for cluster in template_miner.drain.clusters:
        cluster_id = cluster.cluster_id
        template = cluster.get_template()
        templates_info[cluster_id] = {
            "template": template,
            "cluster_id": cluster_id,
            "size": cluster.size,
        }

    # Track template frequencies across time windows
    pattern_trends = defaultdict(list)
    pattern_counts = defaultdict(list)
    all_templates = [info["template"] for info in templates_info.values()]

    # Calculate frequencies for each pattern across time windows
    for window_time, logs in sorted(time_windows.items()):
        if not logs:
            continue

        # Count occurrences of each template in this window
        template_counts = defaultdict(int)

        for log in logs:
            result = template_miner.match(log)
            if result is not None:
                template = result.get_template()
                template_counts[template] += 1

        # Calculate frequencies
        total_logs = len(logs)
        for template in all_templates:
            count = template_counts.get(template, 0)
            frequency = count / total_logs if total_logs > 0 else 0
            pattern_trends[template].append((window_time, frequency))
            pattern_counts[template].append((window_time, count))

    return pattern_trends, pattern_counts


def detect_significant_patterns(
    pattern_trends: dict[str, list[tuple[datetime, float]]],
    threshold: float = 0.2,
    min_occurrences: int = 3,
    focus_on_increases: bool = True,
) -> dict[str, dict[str, Any]]:
    """Identify patterns with significant frequency changes, focusing on increases

    Args:
        pattern_trends: Dictionary mapping patterns to lists of (time, frequency) tuples
        threshold: Minimum relative change to consider significant
        min_occurrences: Minimum number of occurrences required
        focus_on_increases: If True, only detect patterns that increase at some point

    Returns:
        Dictionary of significant patterns with their info
    """
    significant_patterns = {}

    for pattern, frequencies in pattern_trends.items():
        if len(frequencies) < min_occurrences:
            continue

        times, freqs = zip(*frequencies, strict=True)
        freqs = np.array(freqs)

        if np.max(freqs) == 0:
            continue

        # Calculate increase/decrease
        try:
            max_freq = np.max(freqs)
            max_idx = np.argmax(freqs)

            nonzero_indices = np.where(freqs > 0)[0]
            if len(nonzero_indices) == 0:
                continue

            start_idx = nonzero_indices[0]
            start_freq = freqs[start_idx]

            # Calculate increase to max (to detect patterns that increase then decrease)
            max_increase = (max_freq - start_freq) / start_freq if start_freq > 0 else float("inf")

            # Calculate end-to-end change (for full trend direction)
            end_change = (freqs[-1] - start_freq) / start_freq if start_freq > 0 else float("inf")

            increases_at_some_point = max_increase > threshold

            if increases_at_some_point or not focus_on_increases:
                significant_patterns[pattern] = {
                    "peak_increase": float(max_increase),
                    "end_to_end_change": float(end_change),
                    "peak_time": times[max_idx].isoformat()
                    if isinstance(times[max_idx], datetime)
                    else times[max_idx],
                    "frequencies": [
                        (
                            t.isoformat() if isinstance(t, datetime) else t,
                            float(f),
                        )
                        for t, f in frequencies
                    ],
                }
        except Exception as e:
            print(f"Error analyzing pattern: {pattern[:50]}... - {e}")

    return significant_patterns


def find_pattern_examples(
    log_entries: list[LogLine],
    pattern: str,
    template_miner: TemplateMiner,
    max_examples: int = 3,
) -> list[str]:
    examples = []

    for entry in log_entries:
        result = template_miner.match(entry.log)
        if result is not None and result.get_template() == pattern:
            examples.append(entry.log)
            if len(examples) >= max_examples:
                break

    return examples


def detect_log_anomalies(
    log_entries: list[LogLine],
    window_minutes: int = 5,
    threshold: float = 0.2,
    min_occurrences: int = 3,
) -> list[LogAnomaly]:
    if len(log_entries) == 0:
        return []

    config = TemplateMinerConfig()
    config.drain_sim_th = 0.5
    config.drain_depth = 4
    config.masking_instructions = []
    template_miner = TemplateMiner(config=config)

    # Pre-process all logs to train the template miner. If log entries are too
    # big, sample them for training.
    training_sample_rate = 1.0
    if len(log_entries) > MAX_TRAINING_SAMPLE_SIZE:
        training_sample_rate = MAX_TRAINING_SAMPLE_SIZE / len(log_entries)

    for entry in log_entries:
        if random.random() < training_sample_rate:  # noqa: S311
            template_miner.add_log_message(entry.log)

    # Also sample the log entries for analysis
    analysis_sample_rate = 1.0
    if len(log_entries) > MAX_ANALYSIS_SAMPLE_SIZE:
        analysis_sample_rate = MAX_ANALYSIS_SAMPLE_SIZE / len(log_entries)

    analysis_log_entries = [
        entry
        for entry in log_entries
        if random.random() < analysis_sample_rate  # noqa: S311
    ]

    time_windows = segment_logs_by_time(analysis_log_entries, window_minutes=window_minutes)
    pattern_trends, pattern_counts = analyze_pattern_trends(time_windows, template_miner)

    significant_patterns = detect_significant_patterns(
        pattern_trends,
        threshold=threshold,
        min_occurrences=min_occurrences,
        focus_on_increases=True,
    )

    results = []
    for pattern, info in significant_patterns.items():
        results.append(
            LogAnomaly(
                pattern=pattern,
                peak_increase=info["peak_increase"],
                peak_time=info["peak_time"],
                counts_sparkline="".join(
                    sparklines([float(f) for _, f in pattern_counts[pattern]])
                ),
            )
        )

    results.sort(key=lambda x: x.peak_increase, reverse=True)
    return results[:MAX_ANOMALIES_TO_RETURN]



from pathlib import Path
from typing import Any

import typer
import yaml
from pydantic import BaseModel, Field
from pydantic_yaml import to_yaml_str

CONFIG_ROOT = Path(typer.get_app_dir("daydream", force_posix=True))
DEFAULT_CONFIG_FILE = Path(__file__).parent / "defaults.yaml"


type PluginSettings = dict[str, Any]
"""type alias to dict[str, Any] that contains settings for the plugin"""


class PluginConfig(BaseModel):
    enabled: bool = True
    settings: PluginSettings = Field(default_factory=dict)

    def __hash__(self) -> int:
        return hash(self.model_dump_json())


class Config(BaseModel):
    plugins: dict[str, PluginConfig] = {}


DEFAULT_CONFIG = Config(
    **yaml.safe_load(DEFAULT_CONFIG_FILE.read_text()),
)


def get_config_dir(profile: str, create: bool = False) -> Path:
    config_dir = CONFIG_ROOT / "profiles" / profile

    # Create a new config directory if it doesn't exist.
    if not config_dir.exists():
        if not create:
            raise FileNotFoundError(f"Config directory {config_dir} does not exist")
        config_dir.mkdir(parents=True, exist_ok=True)

    return config_dir


def get_config_path(profile: str, create: bool = False) -> Path:
    return get_config_dir(profile, create) / "config.yaml"


def load_config(profile: str, create: bool = False) -> Config:
    config_path = get_config_path(profile, create)

    # Create a new config file if it doesn't exist.
    if not config_path.exists():
        if not create:
            raise FileNotFoundError(f"Config file {config_path} does not exist")
        config_path.write_text(DEFAULT_CONFIG_FILE.read_text())

    # Merge the default config with the user config, with the user config taking precedence.
    return Config(**{**DEFAULT_CONFIG.model_dump(), **yaml.safe_load(config_path.read_text())})


def save_config(cfg: Config, profile: str, create: bool = False) -> None:
    config_path = get_config_path(profile, create)
    yaml_str = to_yaml_str(cfg, default_flow_style=False)
    config_path.write_text(yaml_str, encoding="utf-8")



import asyncio
import sys
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from pathlib import Path
from typing import Any

import uvicorn
from fastapi import FastAPI, Request
from pydantic_settings import BaseSettings
from rich import print
from uvicorn.supervisors import ChangeReload, Multiprocess

from daydream.agent.analysis import AnalysisAgent
from daydream.config.utils import get_config_dir


class Settings(BaseSettings):
    NGROK_TOKEN: str = ""
    NGROK_DOMAIN: str = ""

    DAYDREAM_TUNNEL: bool = False
    DAYDREAM_RELOAD: bool = False
    DAYDREAM_PROFILE: str = "default"
    DAYDREAM_HOST: str = "127.0.0.1"
    DAYDREAM_PORT: int = 8000
    DAYDREAM_WORKERS: int = 1


settings = Settings()


@asynccontextmanager
async def lifespan(app: FastAPI) -> AsyncGenerator[None, None]:
    tunnel = settings.DAYDREAM_TUNNEL
    ngrok_domain = settings.NGROK_DOMAIN
    ngrok_token = settings.NGROK_TOKEN
    host = settings.DAYDREAM_HOST
    port = settings.DAYDREAM_PORT
    internal_url = f"http://{host}:{port}/webhook"
    external_url = None

    if tunnel:
        from pyngrok import ngrok

        ngrok.set_auth_token(ngrok_token)
        listener = ngrok.connect(
            addr=f"{host}:{port}", name=f"daydream-{port}", domain=ngrok_domain
        )
        external_url = f"{listener.public_url}/webhook"
        print(f"Tunnel opened: {external_url} -> {internal_url}")

        try:
            yield
        finally:
            print(f"Closing tunnel {external_url}")
            ngrok.kill()
    else:
        yield


app = FastAPI(lifespan=lifespan)
app.state.analysis_tasks = set()


async def analyze(payload: dict[str, Any], profile: str) -> None:
    analysis_agent = AnalysisAgent(profile)
    analysis = await analysis_agent.acall(payload=payload)
    print(analysis)


@app.post("/webhook")
async def webhook(request: Request) -> dict[str, Any]:
    analysis_tasks = request.app.state.analysis_tasks

    # Extract the alert payload from the request.
    payload = await request.json()

    # Start a background task to analyze the payload.
    task = asyncio.create_task(analyze(payload, settings.DAYDREAM_PROFILE))

    # Add task to the set to create a strong reference.
    analysis_tasks.add(task)

    # Remove the task from the set when it completes to avoid memory leaks.
    task.add_done_callback(analysis_tasks.discard)

    return {"message": "OK"}


async def listen(
    host: str = "127.0.0.1",
    port: int = 8000,
    workers: int = 1,
    profile: str = "default",
    reload: bool = False,
    tunnel: bool = False,
    ngrok_token: str = "",
    ngrok_domain: str = "",
) -> None:
    # Set environment variables to make them accessible to FastAPI's lifespan.
    settings.DAYDREAM_PROFILE = profile
    settings.DAYDREAM_TUNNEL = tunnel
    settings.NGROK_TOKEN = ngrok_token
    settings.NGROK_DOMAIN = ngrok_domain
    settings.DAYDREAM_HOST = host
    settings.DAYDREAM_PORT = port

    # Recurse up the directory tree until we find the pyproject.toml file (i.e.,
    # the project root).
    reload_config = {}
    if reload:
        project_root = Path(__file__).parent
        while "__version__" not in (project_root / "__init__.py").read_text():
            project_root = project_root.parent
            if not project_root.exists():
                raise FileNotFoundError("Could not find the project root")
        reload_config = {
            "reload": True,
            "reload_dirs": [str(project_root), str(get_config_dir(profile))],
            "reload_includes": ["**/*.py", "**/*.yaml"],
        }

    uvicorn_config = uvicorn.Config(
        f"{__name__}:app",
        host=host,
        port=port,
        workers=workers,
        ws="none",
        **reload_config,
    )

    server = uvicorn.Server(config=uvicorn_config)

    try:
        if uvicorn_config.should_reload:
            sock = uvicorn_config.bind_socket()
            ChangeReload(uvicorn_config, target=server.run, sockets=[sock]).run()
        elif uvicorn_config.workers > 1:
            sock = uvicorn_config.bind_socket()
            Multiprocess(uvicorn_config, target=server.run, sockets=[sock]).run()
        else:
            await server.serve()
    except KeyboardInterrupt:
        pass  # pragma: full coverage
    finally:
        uds = Path(uvicorn_config.uds) if uvicorn_config.uds else None
        if uds and uds.exists():
            uds.unlink()  # pragma: py-win32

    if not server.started and not uvicorn_config.should_reload and uvicorn_config.workers == 1:
        sys.exit(3)



import asyncio
from collections import defaultdict
from datetime import datetime
from typing import TYPE_CHECKING, Any

from aioboto3 import Session
from pydantic import BaseModel, Field

from daydream.knowledge import Node
from daydream.models import Observation
from daydream.plugins.aws.utils import (
    ensure_aws_session,
    list_accessible_regions_for_service,
    swallow_boto_client_access_errors,
)

if TYPE_CHECKING:
    from pydantic import AwareDatetime


CLOUDWATCH_DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ"


DEFAULT_AWS_ACCOUNT_NAME = "default"


class AwsAccount(BaseModel):
    name: str | None = Field(default=DEFAULT_AWS_ACCOUNT_NAME)
    profile: str | None = Field(default=None)

    @property
    def session(self) -> Session:
        if not hasattr(self, "_session"):
            self._session = Session(profile_name=self.profile) if self.profile else Session()
        return self._session


class AwsNode(Node):
    """A base class for all AWS nodes."""

    aws_account: AwsAccount = Field()

    @property
    def session(self) -> Session:
        if not hasattr(self, "_session"):
            self._session = self.aws_account.session
        return self._session

    async def _get_cloudwatch_metric_for_accessible_regions(
        self,
        **cloudwatch_params: Any,
    ) -> list[Observation]:
        await ensure_aws_session(self.session)
        accessible_regions = await list_accessible_regions_for_service(self.session, "cloudwatch")

        # We don't know which region a given resource is in, so check all regions
        # in parallel.
        observation_results = await asyncio.gather(
            *(
                self._get_cloudwatch_metric_for_region(region, **cloudwatch_params)
                for region in accessible_regions
            )
        )

        # Flatten into a list of observations
        observations = []
        [observations.extend(result) for result in observation_results]
        return observations

    async def _get_cloudwatch_metric_for_region(
        self,
        region: str,
        **cloudwatch_params: Any,
    ) -> list[Observation]:
        metric_name = cloudwatch_params.get("MetricName", "UnknownMetric")
        async with (
            swallow_boto_client_access_errors(service_name="cloudwatch", region=region),
            self.session.client("cloudwatch", region_name=region) as client,
        ):
            result = await client.get_metric_statistics(**cloudwatch_params)
            observations: list[Observation] = []
            datapoints: list[dict[str, Any]] = result.get("Datapoints", [])
            if datapoints:
                series: defaultdict[str, dict[AwareDatetime, float]] = defaultdict(dict)

                for point in datapoints:
                    dt = point.get("Timestamp")
                    if not isinstance(dt, datetime):
                        raise ValueError(f"Timestamp is not a datetime: {dt}")

                    # Each point can have multiple values - Average, Minimum,
                    # Maximum, etc. Create a series for each.
                    for key, value in point.items():
                        if key not in ["Timestamp", "Unit"]:
                            series[key][dt] = float(value)

                # Now that we have collected the values for each series,
                # convert them into Observations.
                for key, data in series.items():
                    observations.append(
                        Observation(
                            node_id=self.nid,
                            observation_type=f"{key} {metric_name}",
                            data=data,
                        )
                    )

            return observations



from collections.abc import AsyncGenerator

import pytest_asyncio
from fastmcp import Client, FastMCP

from daydream.config.utils import Config
from daydream.knowledge import Graph
from daydream.mcp import Context, build_mcp_server
from daydream.plugins import PluginManager


@pytest_asyncio.fixture
async def default_config() -> Config:
    return PluginManager.default_config


@pytest_asyncio.fixture
async def mcp_server(default_config: Config) -> FastMCP:
    plugins = PluginManager(config=default_config)
    server = await build_mcp_server(
        Context(
            profile="test",
            config=default_config,
            plugins=plugins,
            graph=Graph(),
        )
    )

    return server


@pytest_asyncio.fixture
async def mcp_client(mcp_server: FastMCP) -> AsyncGenerator[Client, None]:
    async with Client(mcp_server) as client:
        yield client



from daydream.knowledge import Node
from daydream.utils import import_submodules

from .base import REGISTRY, Plugin, PluginCapability, PluginManager

# Autoload all plugins in the plugins directory
import_submodules("daydream.plugins")

__all__ = ["REGISTRY", "Node", "Plugin", "PluginCapability", "PluginManager"]



from pydantic import AwareDatetime

from daydream.knowledge import HasMetrics
from daydream.models import Observation
from daydream.plugins.aws.nodes.base import AwsNode


class AwsEc2Instance(AwsNode, HasMetrics):
    """An EC2 instance."""

    async def get_identifiers(self) -> list[str | None]:
        return [
            *await super().get_identifiers(),
            self.raw_data["InstanceId"],
            self.raw_data.get("PrivateIpAddress"),
            self.raw_data.get("PublicIpAddress"),
            self.raw_data.get("PrivateDnsName"),
            self.raw_data.get("PublicDnsName"),
        ]

    async def get_reference_identifiers(
        self,
    ) -> list[str | None | tuple[str | None, str]]:
        return [
            *await super().get_reference_identifiers(),
            *(sg["GroupId"] for sg in self.raw_data["SecurityGroups"]),
            *((m["Ebs"]["VolumeId"], "has_volume") for m in self.raw_data["BlockDeviceMappings"]),
            self.raw_data.get("VpcId"),
            self.raw_data.get("SubnetId"),
        ]

    async def list_available_metrics(self) -> list[str]:
        return [
            "CPUUtilization",
            "NetworkIn",
            "NetworkOut",
            "StatusCheckFailed",
            "EBSReadBytes",
            "EBSWriteBytes",
        ]

    async def get_metric(
        self,
        metric_name: str,
        time_range_start: AwareDatetime,
        time_range_end: AwareDatetime,
    ) -> list[Observation] | str:
        return await self._get_cloudwatch_metric_for_accessible_regions(
            Namespace="AWS/EC2",
            MetricName=metric_name,
            Dimensions=[{"Name": "InstanceId", "Value": self.raw_data["InstanceId"]}],
            StartTime=time_range_start,
            EndTime=time_range_end,
            Period=300,
            Statistics=["Average", "Minimum", "Maximum"],
        )



import time
from collections import Counter

import anyio
import typer

from daydream.cli._app import app
from daydream.cli.options import PROFILE_OPTION
from daydream.config import load_config
from daydream.config.utils import get_config_dir
from daydream.knowledge import Graph
from daydream.plugins import PluginManager
from daydream.plugins.mixins import KnowledgeGraphMixin
from daydream.telemetry import client as telemetry


@app.command()
def build_graph(
    profile: str = PROFILE_OPTION,
    interval: int | None = typer.Option(
        None,
        "--interval",
        help="Rebuild the graph continuously, pausing for the specified interval between builds",
    ),
) -> None:
    """Build a knowledge graph for your cloud infrastructure"""

    async def _build_graph() -> None:
        print("Building graph...")

        start_time = time.perf_counter()

        graph = Graph()
        config = load_config(profile, create=True)
        output_path = (get_config_dir(profile) / "graph.json").resolve()
        plugin_manager = PluginManager(config)

        async with anyio.create_task_group() as tg:
            for plugin in plugin_manager.get_plugins_with_capability(KnowledgeGraphMixin):
                print(f"Populating graph with the {plugin.name} plugin...")
                tg.start_soon(plugin.populate_graph, graph)

        await graph.infer_edges()

        print(f"Saving graph to {output_path!s}...")
        await graph.save(output_path)

        end_time = time.perf_counter()
        total_time = end_time - start_time
        edge_counts = Counter(
            [edge.properties["relationship_type"] async for edge in graph.iter_edges()]
        )
        node_counts = Counter([node.node_type async for node in graph.iter_nodes()])

        print("=== Summary ===")

        print("Edges:")
        for relationship_type, count in edge_counts.items():
            print(f"  {relationship_type}: {count}")
        print("Nodes:")
        for node_type, count in node_counts.items():
            print(f"  {node_type}: {count}")

        print(f"Graph built in {total_time:.2f} seconds")
        print(f"Graph saved to {output_path!s}")
        print("=== End Summary ===")

        await telemetry.send_event(
            {
                "command": "build_graph",
                "profile": profile,
                "duration_seconds": total_time,
                "node_counts": node_counts,
                "edge_counts": edge_counts,
            }
        )

    if interval:
        while True:
            anyio.run(_build_graph)
            print(f"Sleeping for {interval} seconds before next build...")
            time.sleep(interval)
    else:
        anyio.run(_build_graph)



import os
import sys
import traceback
import warnings
from typing import TextIO


def warn_with_traceback(
    message: str,
    category: type[Warning],
    filename: str,
    lineno: int,
    file: TextIO | None = None,
    line: str | None = None,
) -> None:
    log = file if file and hasattr(file, "write") else sys.stderr
    traceback.print_stack(file=log)
    log.write(warnings.formatwarning(message, category, filename, lineno, line))


# Render warnings with traceback to make it easier to track down the culprit.
if os.getenv("WARN_WITH_TRACEBACK", "false").lower() == "true":
    warnings.showwarning = warn_with_traceback



import json
import shlex
import shutil
from pathlib import Path
from typing import Annotated, Any

import anyio
import rich
import typer
from fastmcp.cli.claude import get_claude_config_path

from daydream.cli._app import app
from daydream.cli.options import DEFAULT_PROFILE, PROFILE_OPTION
from daydream.config.utils import Config, save_config
from daydream.plugins.base import PluginManager
from daydream.utils import Choice, checkbox, confirm, iprint


def _resolve_default_claude_mcp_config_filename() -> Path | None:
    claude_config_dir = get_claude_config_path()
    if claude_config_dir is None:
        return None
    return claude_config_dir / "claude_desktop_config.json"


_default_claude_mcp_config_filename = _resolve_default_claude_mcp_config_filename()


def _resolve_default_uvx() -> Path:
    uv_path = shutil.which("uvx")
    if uv_path:
        return Path(uv_path).absolute()
    return Path("uvx")


_default_uvx_path = _resolve_default_uvx()


def _resolve_default_uv() -> Path:
    uv_path = shutil.which("uv")
    if uv_path:
        return Path(uv_path).absolute()
    return Path("uv")


_default_uv_path = _resolve_default_uv()


def _resolve_default_use_uv_run() -> bool:
    try:
        cwd = Path.cwd()
        git_dir = None
        for parent in [cwd, *cwd.parents[:10]]:
            if (parent / ".git").exists():
                git_dir = parent
                break
        if git_dir is None:
            return False
        repo_name = git_dir.name
        return repo_name == "daydream"
    except Exception:
        return False


_default_use_uv_run = _resolve_default_use_uv_run()


def _resolve_default_repo_path() -> Path:
    return Path.cwd().absolute()


_default_repo_path = _resolve_default_repo_path()


def _resolve_default_claude_app() -> Path:
    return Path("/Applications/Claude.app").resolve()


_default_claude_app = _resolve_default_claude_app()


@app.command()
def configure(
    profile: str = PROFILE_OPTION,
    uvx_path: Annotated[
        Path,
        typer.Option(
            ...,
            "--uvx-path",
            help="Path to the uvx executable",
            show_default=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            writable=False,
            resolve_path=False,
            exists=True,
        ),
    ] = _default_uvx_path,
    use_uv_run: bool = typer.Option(
        _default_use_uv_run,
        "--use-uv-run",
        help="Use uv run instead of uvx to start the Daydream MCP server (useful for develping Daydream)",
    ),
    uv_path: Annotated[
        Path,
        typer.Option(
            ...,
            "--uv-path",
            help="Path to the uv executable",
            show_default=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            writable=False,
            resolve_path=False,
            exists=False,
        ),
    ] = _default_uv_path,
    mcp_command_prefix: Annotated[
        str | None,
        typer.Option(
            ...,
            "--mcp-command-prefix",
            help="Command prefix to use for the Daydream MCP Server (e.g., 'aws-vault exec your_aws_profile')",
        ),
    ] = None,
    mcp_command_directory: Annotated[
        Path | None,
        typer.Option(
            ...,
            "--mcp-command-dir",
            help="Path to the directory MCP server will be started",
            show_default=True,
            file_okay=False,
            dir_okay=True,
            readable=True,
            writable=False,
            resolve_path=False,
            exists=True,
        ),
    ] = _default_repo_path,
    claude_app: Annotated[
        Path | None,
        typer.Option(
            ...,
            "--claude-app",
            help="Path to the Claude Desktop app",
            show_default=True,
            file_okay=False,
            dir_okay=True,
            readable=True,
            writable=False,
            resolve_path=False,
            exists=False,
        ),
    ] = _default_claude_app,
    claude_config_file: Annotated[
        Path | None,
        typer.Option(
            ...,
            "--claude-config-filename",
            help="Path to the Claude Desktop MCP config json file",
            show_default=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            writable=True,
            resolve_path=False,
            exists=False,
        ),
    ] = _default_claude_mcp_config_filename,
) -> None:
    async def _recipe() -> None:
        _welcome_to_daydream()
        cfg = PluginManager.default_config
        plugin_mgr = PluginManager(cfg)
        await _select_plugins_to_enable_disable(cfg)
        iprint("")
        await _configure_plugins(cfg, plugin_mgr)
        iprint("")
        await _validate_plugins(cfg, plugin_mgr)
        save_config(cfg, profile, create=True)
        await _suggest_building_graph(profile, use_uv_run)
        daydream_mcp_server_config = _generate_mcp_server_config(
            uvx_path, use_uv_run, uv_path, mcp_command_prefix, mcp_command_directory, profile
        )
        await _setup_clients_and_open(daydream_mcp_server_config, claude_app, claude_config_file)

    anyio.run(_recipe)


def _welcome_to_daydream() -> None:
    rich.print("""[red]
                ⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⢀⣀⠤⠤⠒⠒⠒⠒⠒⠒⠤⢤⣀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀[red]
                ⠀⠀⠀⠀⠀⠀⠀⠀⠀⣠⠴⠚⠁⠀⠀⠀⠀⣀⣀⣀⡀⠀⠀⠀⠀⠙⠲⢄⠀⠀⠀⠀⠀⠀⠀⠀[orange]
                ⠀⠀⠀⠀⠀⠀⢀⡴⠊⠀⠀⢀⡠⠖⠒⠉⠉⠀⠀⠀⠈⠉⠑⠲⢤⡀⠀⠀⠑⣄⠀⠀⠀⠀⠀⠀[yellow]
                ⠀⠀⠀⠀⠀⣠⠋⠀⠀⣠⠴⠋⠀⠀⢀⡠⠤⠔⠒⠒⠦⠤⣀⠀⠀⠙⢦⡀⠀⠈⢣⠀⠀⠀⠀⠀[yellow]
                ⠀⠀⠀⠀⡰⠁⠀⠀⡼⠋⠀⠀⡠⠚⠁⠀⠀⠀⠀⠀⠀⠀⠈⠑⢄⠀⠀⠱⡄⠀⠀⢳⠀⠀⠀⠀[green]
                ⠀⠀⠀⢰⠁⠀⠀⡞⠀⠀⣠⠊⠀⠀⣠⠔⠊⠉⠉⠁⠒⠤⡀⠀⠈⢳⡀⠀⢳⡀⠀⠀⢧⠀⠀⠀[green]
                ⠀⠀⠀⡇⠀⠀⡸⠁⠀⡰⠁⠀⢀⡜⠁⠀⠀⠀⠀⠀⠀⠀⠈⢢⠀⠐⢧⠀⠀⢧⠀⠀⠘⡄⠀⠀[blue]
                ⠀⠀⢸⠁⠀⣠⡇⢀⣠⣇⡀⢀⠎⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⣇⠀⣸⣄⣀⠘⣆⡀⠀⣇⠀⠀[blue]
                ⠀⠀⢸⣞⠉⠈⠙⡟⠀⠀⣻⠞⠳⣄⠀⠀⠀⠀⠀⠀⠀⠀⢠⡴⠛⢾⠁⠀⢘⡿⠁⠈⠹⣿⠀⠀[magenta]
                ⠀⣴⠟⠿⠀⠀⠀⠀⠀⠀⠀⠀⢠⠾⣄⠀⠀⠀⠀⠀⠀⢀⠼⣅⡀⠀⠀⠀⠀⠀⠀⠀⠘⠛⢦⠀[magenta]
                ⠸⣇⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⢀⡀⣸⡇⠀⠀⠀⠀⠀⡏⠀⢁⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⣠⡇[bright_magenta]
                ⠀⠈⠹⠃⠀⠀⡄⠀⠀⠠⡀⠀⠙⣿⠉⠀⠀⠀⠀⠀⠀⠈⢹⠋⠀⠀⡤⠀⠀⢰⡄⠀⠀⢻⠉⠀[bright_magenta]
                ⠀⠀⠘⠦⠤⠼⠳⢄⣀⡰⠓⠂⠐⠋⠀⠀⠀⠀⠀⠀⠀⠀⠈⠓⠀⠚⠧⣀⣠⠞⠳⠤⠤⠟⠁⠀[default]

    ██████╗  █████╗ ██╗   ██╗██████╗ ██████╗ ███████╗ █████╗ ███╗   ███╗
    ██╔══██╗██╔══██╗╚██╗ ██╔╝██╔══██╗██╔══██╗██╔════╝██╔══██╗████╗ ████║
    ██║  ██║███████║ ╚████╔╝ ██║  ██║██████╔╝█████╗  ███████║██╔████╔██║
    ██║  ██║██╔══██║  ╚██╔╝  ██║  ██║██╔══██╗██╔══╝  ██╔══██║██║╚██╔╝██║
    ██████╔╝██║  ██║   ██║   ██████╔╝██║  ██║███████╗██║  ██║██║ ╚═╝ ██║
    ╚═════╝ ╚═╝  ╚═╝   ╚═╝   ╚═════╝ ╚═╝  ╚═╝╚══════╝╚═╝  ╚═╝╚═╝     ╚═╝

> Welcome to Daydream!
>
> This interactive tool will walk through the configuration of your system.
> We'll setup Daydream plugins, configuration, and configure clients like Claude Desktop or Amazon Q to use the Daydream MCP server...
> """)


async def _select_plugins_to_enable_disable(
    cfg: Config,
) -> None:
    iprint("Daydream uses plugins to access your infrastructure resources, like AWS or Aptible.")
    iprint("Each plugin may be enabled or disabled to match your desired knowledge graph")
    iprint(
        "Some plugins require configuration, for example the AWS plugin needs access to your AWS account"
    )
    iprint("")
    iprint("These are the plugins we recommend:")
    for (
        plugin_name,
        plugin_cfg,
    ) in cfg.plugins.items():
        if plugin_cfg.enabled:
            iprint(f"• {plugin_name}")
    iprint("")
    if await confirm(
        "Would you like to change this?",
        default=False,
        auto_enter=False,
    ):
        await _enable_disable_plugins(cfg)


async def _enable_disable_plugins(cfg: Config) -> None:
    possible_plugins = list(cfg.plugins.keys())
    enabled_search = len(possible_plugins) > 10
    selected_plugins = await checkbox(
        "Which plugins would you like to enable for the Daydream MCP Server? Press enter to continue",
        choices=[Choice(p, checked=plugin_cfg.enabled) for p, plugin_cfg in cfg.plugins.items()],
        use_search_filter=enabled_search,
        use_jk_keys=not enabled_search,
    )
    for plugin_name, plugin_cfg in cfg.plugins.items():
        plugin_cfg.enabled = plugin_name in selected_plugins


async def _configure_plugins(cfg: Config, plugin_mgr: PluginManager) -> None:
    iprint("Now we will go through plugin specific configuration...")
    iprint("")
    for plugin_name, plugin_cfg in cfg.plugins.items():
        if not plugin_cfg.enabled:
            continue
        iprint(f"Configuring [bold]{plugin_name}[/bold] plugin settings...")
        plugin = plugin_mgr.get_plugin(plugin_name)
        plugin_settings = await plugin.interactive_configure(cfg)
        plugin_cfg.settings = plugin_settings
        iprint(f"✅ {plugin_name} plugin settings configured successfully!")
        iprint("")


async def _validate_plugins(cfg: Config, plugin_mgr: PluginManager) -> None:
    iprint(
        "And now we'll verify that these plugins have the appropriate configuration, and validate any authentication works as expected..."
    )
    iprint("")
    for plugin_name, plugin_cfg in cfg.plugins.items():
        if not plugin_cfg.enabled:
            continue
        iprint(f"Validating {plugin_name} plugin...")
        plugin = plugin_mgr.get_plugin(plugin_name, plugin_cfg)
        # FIXME: propagate failed validation... but don't cause the process to stop
        attempt_validation = True
        config_valid = False
        err: Exception | None = None
        while attempt_validation:
            try:
                await plugin.validate_plugin_config()
            except Exception as ex:
                iprint(f"{plugin_name} validation error: {ex!s}")
                attempt_validation = await confirm(f"Retry {plugin_name} validation?", default=True)
                err = ex
            else:
                config_valid = True
                attempt_validation = False
        if config_valid:
            iprint(f"✅ {plugin_name} has valid config and is working!")
        else:
            iprint(f"❌ {plugin_name} validation failed: {err!s}")


async def _suggest_building_graph(profile: str, use_uv_run: bool) -> None:
    iprint("We recommend building the knowledge graph before proceeding.")
    iprint("Try running this command in another window:")
    iprint("")
    iprint(
        f"    {'uvx' if not use_uv_run else 'uv run'} daydream build-graph{f' --profile {profile}' if profile != DEFAULT_PROFILE else ''}"
    )
    iprint("")
    while not await confirm("Ready to proceed?", default=True):
        pass
    iprint("")


def _generate_mcp_server_config(
    uvx_path: Path,
    use_uv_run: bool,
    uv_path: Path,
    mcp_command_prefix: str | None,
    mcp_command_directory: Path | None,
    profile: str,
) -> dict[str, Any]:
    command, *args = [
        *shlex.split(mcp_command_prefix or ""),
        *([str(uvx_path)] if not use_uv_run else [str(uv_path), "run"]),
        *(["--directory", str(mcp_command_directory)] if mcp_command_directory else []),
        "daydream",
        "start",
        "--profile",
        profile,
        "--disable-http",
    ]
    env = {
        "HOME": str(Path.home()),
    }
    return {
        "command": shutil.which(command),
        "args": args,
        "env": env,
    }


async def _setup_clients_and_open(
    daydream_mcp_server_config: dict[str, Any],
    claude_app: Path | None,
    claude_config_file: Path | None,
) -> None:
    # claude code: ~/Library/Application Support/Claude/claude_desktop_config.json
    # amazon q: ~/.aws/amazonq/mcp.json
    # cursor: ~/.cursor/mcp.json
    # windsurf: ~/.codeium/windsurf/mcp_config.json
    if not await confirm(
        "Would you like to setup Claude Desktop to use the Daydream MCP server?",
        default=True,
    ):
        return
    if claude_app is None or not claude_app.exists():
        iprint(f"❌ Claude Desktop app not found at [bright_black]{claude_app}[/bright_black]")
        iprint(
            "Please let us know about this issue, and try again with the --claude-app <path> flag"
        )
        iprint("")
        return
    if claude_config_file is None:
        iprint(
            f"❌ Claude Desktop MCP configuration file not found at [bright_black]{claude_config_file}[/bright_black]"
        )
        iprint(
            "Please let us know about this issue, and try again with the --claude-config-filename <path> flag"
        )
        iprint("")
        return
    await _setup_claude_desktop_client(daydream_mcp_server_config, claude_app, claude_config_file)


async def _merge_daydream_mcp_config(
    daydream_mcp_server_config: dict[str, Any], config_file: Path
) -> dict[str, Any]:
    try:
        mcp_config = json.loads(config_file.read_text())
    except (json.JSONDecodeError, FileNotFoundError):
        mcp_config = {}

    return {
        **mcp_config,
        "mcpServers": {
            **mcp_config.get("mcpServers", {}),
            "daydream": daydream_mcp_server_config,
        },
    }


async def _setup_claude_desktop_client(
    daydream_mcp_server_config: dict[str, Any], claude_app: Path, claude_config_file: Path
) -> None:
    mcp_servers = await _merge_daydream_mcp_config(daydream_mcp_server_config, claude_config_file)
    claude_config_file.write_text(json.dumps(mcp_servers, indent=2))
    iprint(
        f"✅ Updated Claude Desktop MCP config file at [bright_black]{claude_config_file.resolve()!s}[/bright_black]"
    )
    if await confirm("Would you like to open the Claude Desktop app now?", default=True):
        iprint(f"Opening [bright_black]{claude_app}[/bright_black]...")
        typer.launch(str(claude_app))
    iprint("")



import sys
import warnings

import anyio
import typer
from fastmcp import settings as fastmcp_settings

from daydream import mcp
from daydream.cli._app import app
from daydream.cli.options import PROFILE_OPTION
from daydream.telemetry import client as telemetry


@app.command()
def start(
    profile: str = PROFILE_OPTION,
    disable_sse: bool = typer.Option(
        False,
        "--disable-sse",
        help="Disable the HTTP transport for the MCP Server (deprecated, use --disable-http instead)",
    ),
    disable_stdio: bool = typer.Option(
        False, "--disable-stdio", help="Disable the stdio transport for the MCP Server"
    ),
    disable_http: bool = typer.Option(
        False, "--disable-http", help="Disable the HTTP transport for the MCP Server"
    ),
    http_host: str = typer.Option(
        fastmcp_settings.host, "--http-host", help="The host to bind the HTTP transport to"
    ),
    http_port: int = typer.Option(
        fastmcp_settings.port, "--http-port", help="The port to bind the HTTP transport to"
    ),
) -> None:
    """Start the Daydream MCP Server"""
    # Deprecate --disable-sse in favor of --disable-http
    if "--disable-sse" in sys.argv:
        disable_http = disable_sse
        warnings.warn(
            "The `--disable-sse` argument is deprecated. Use `--disable-http` instead.",
            DeprecationWarning,
            stacklevel=2,
        )

    async def _start() -> None:
        await telemetry.send_event(
            {
                "command": "start",
                "profile": profile,
                "disable_sse": disable_sse,
                "disable_stdio": disable_stdio,
                "disable_http": disable_http,
                "http_host": http_host,
                "http_port": http_port,
            }
        )
        await mcp.start(
            profile=profile,
            disable_stdio=disable_stdio,
            disable_http=disable_http,
            http_host=http_host,
            http_port=http_port,
        )

    anyio.run(_start)



from collections.abc import Iterator
from typing import TYPE_CHECKING, Any, Protocol, TypeVar, cast

from daydream.config.utils import Config, PluginConfig, PluginSettings
from daydream.utils import classproperty

if TYPE_CHECKING:
    from daydream.mcp import Context

REGISTRY: dict[str, type["Plugin"]] = {}


class PluginProtocol(Protocol):
    """Protocol for the plugins interface."""

    context: "Context"

    @classproperty
    def name(cls) -> str:
        """Converts a SomePlugin class to a "some" string, which is the key used in the configuration file.

        For example, AwsPlugin becomes "aws".
        """
        return cast("type[PluginProtocol]", cls).__name__.removesuffix("Plugin").lower()


class Plugin(PluginProtocol):
    """Base class for all plugins."""

    def __init__(self, **settings: Any) -> None:
        self._settings = settings

    def __init_subclass__(cls, **kwargs: Any) -> None:
        REGISTRY[cls.name] = cls

    @classproperty
    def default_plugin_settings(cls) -> PluginSettings:
        return {}

    def init_plugin(self) -> None:
        pass

    async def validate_plugin_config(self) -> None:
        """Validate the plugin configuration, ensure authentication works, and anything else that helps ensure the plugin will work when used.

        Plugins should raise an exception if the configuration is invalid or unusable.

        This is called from the `daydream configure` command.
        """
        pass

    async def interactive_configure(self, cfg: Config) -> PluginSettings:
        """Interactive wizard for configuring the settings of this plugin. Return final PluginSettings here. Caller is responsible for updating the associated PluginConfig.

        This is called from the `daydream configure` command.
        """
        return {}


class PluginCapability(PluginProtocol):
    """Protocol for capabilities that plugins can have."""


HasPluginCapability = TypeVar("HasPluginCapability", bound=PluginCapability)


class PluginManager:
    """A manager for plugins and their configuration."""

    def __init__(self, config: Config) -> None:
        self._config = config
        self._plugins: dict[tuple[str, PluginConfig], Plugin] = {}

    def get_plugin_class(self, name: str) -> type[Plugin]:
        """Return the plugin class for a given name."""
        return REGISTRY[name]

    def get_plugin(self, name: str, config: PluginConfig | None = None) -> Plugin:
        """Return the configured plugin for a given name."""
        config = config or self._config.plugins[name]
        key = (name, config)
        if key not in self._plugins:
            self._plugins[key] = self.get_plugin_class(name)(**config.settings)
            self._plugins[key].init_plugin()
        return self._plugins[key]

    def get_enabled_plugins(self) -> list[Plugin]:
        """Return a list of enabled plugins."""
        return [
            self.get_plugin(name) for name, config in self._config.plugins.items() if config.enabled
        ]

    def get_plugins_with_capability(
        self,
        capability: type[HasPluginCapability],
    ) -> list[HasPluginCapability]:
        """Return a list of plugins that have the given capability."""
        return [plugin for plugin in self.get_enabled_plugins() if isinstance(plugin, capability)]

    def __iter__(self) -> Iterator[Plugin]:
        """Iterate over all enabled plugins."""
        return iter(self.get_enabled_plugins())

    @classproperty
    def default_config(cls) -> Config:
        cfg = Config()
        for p, plugin_cls in REGISTRY.items():
            cfg.plugins[p] = PluginConfig(
                enabled=(p != "logs"), settings=plugin_cls.default_plugin_settings
            )
        return cfg



import json
import os
import uuid
from importlib.metadata import version
from typing import Any

import httpx
import sentry_sdk

from daydream.config.utils import CONFIG_ROOT


def _get_or_create_user_id() -> str:
    identity_file = CONFIG_ROOT / ".identity"
    if not identity_file.exists():
        if not identity_file.parent.exists():
            identity_file.parent.mkdir(parents=True)
        identity_file.write_text(str(uuid.uuid4()))
    return identity_file.read_text().strip()


class TunaClient(httpx.AsyncClient):
    BASE_URL = "https://tuna.aptible.com"

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(
            *args,
            base_url=self.BASE_URL,
            headers={
                "Content-Type": "application/json; charset=utf-8",
            },
            timeout=httpx.Timeout(
                connect=30,
                read=30,
                write=45,
                pool=15,
            ),
            **kwargs,
        )
        self._user_id = _get_or_create_user_id()
        self._run_id = str(uuid.uuid4())

    @property
    def user_id(self) -> str:
        """Unique user_id that is created for each invocation of the daydream program"""
        return self._user_id

    async def send_event(self, event: dict[str, Any]) -> None:
        """Record a telemetry event."""
        try:
            uname = os.uname()
            response = await self.get(
                "/www/e",
                params={
                    "id": str(uuid.uuid4()),
                    "user_id": self.user_id,
                    "type": "daydream_telemetry",
                    "url": "https://github.com/aptible/daydream",
                    "value": json.dumps(
                        {
                            "version": version("daydream"),
                            "github": os.getenv("GITHUB_ACTIONS"),
                            "gitlab": os.getenv("GITLAB_CI"),
                            "travis": os.getenv("TRAVIS"),
                            "circleci": os.getenv("CIRCLECI"),
                            "sysname": uname.sysname,
                            "sysmachine": uname.machine,
                            "sysversion": uname.version,
                            "run_id": self._run_id,
                            **event,
                        }
                    ),
                },
            )
            response.raise_for_status()
        except Exception as e:
            sentry_sdk.capture_exception(e)
            # Don't re-raise the exception (telemetry isn't critical)


client = TunaClient()



from pathlib import Path

from pydantic import AwareDatetime
from pydantic_core import to_json

from daydream.knowledge.nodes.mixins.logs import HasLogs
from daydream.lib.log_anomalies import LogAnomaly, detect_log_anomalies
from daydream.plugins.base import Plugin
from daydream.plugins.mixins import McpServerMixin, tool
from daydream.utils import print


class LogsPlugin(Plugin, McpServerMixin):
    @tool()
    async def get_log_file_anomalies(
        self,
        node_id: str,
        time_range_start: AwareDatetime,
        time_range_end: AwareDatetime,
    ) -> list[LogAnomaly] | str:
        """Get anomalies from the log files of a resource using its node ID.

        Limit the time range to an hour or so.

        Get metrics first. Then, find a time in the metrics where it looks
        like an incident started happening. Call this tool with a time range
        30 mins before through 30 mins after the incident start time.

        This tool will return possible log anomaly patterns, but they may not
        all be relevant (they might not be related to the incident, and they
        might not even be errors).

        When using results from log anomalies, always show the user the
        example log entries so that they can make a judgement call.
        """
        node = await self.context.graph.get_node(node_id)

        if not node:
            return f"Resource with node ID {node_id} not found"

        if not isinstance(node, HasLogs):
            return f"Resource with node ID {node_id} does not support logs"

        print(f"Getting logs for {node.nid}")

        node_logs = await node.get_logs(time_range_start, time_range_end)

        print(f"Found {len(node_logs)} logs")

        log_anomalies = detect_log_anomalies(node_logs, window_minutes=1)

        if log_anomalies:
            anomalies_file = (
                Path.home()
                / ".daydream"
                / "logs"
                / f"{node_id.replace(':', '-')}-{time_range_start.date()}-{time_range_end.date()}.json"
            )
            anomalies_file.write_bytes(to_json(log_anomalies, indent=2))

            return f"""STOP HERE, DO NOT CALL OTHER TOOLS YET.

            Log anomalies written to {anomalies_file}. Ask the user to look through that
            file and make sure there is no sensitive data. If there is nothing
            sensitive, ask the user to paste the anomalies back in to this
            conversation so you can interpret the results.
            """
        else:
            return "No log anomalies found."



from pydantic import AwareDatetime

from daydream.knowledge import HasMetrics
from daydream.models import Observation

from .base import AwsNode


class AwsS3Bucket(AwsNode, HasMetrics):
    """An S3 bucket."""

    async def get_identifiers(self) -> list[str | None]:
        return [
            *await super().get_identifiers(),
            self.raw_data["Name"],
            f"arn:aws:s3:::{self.raw_data['Name']}",
        ]

    async def get_reference_identifiers(
        self,
    ) -> list[str | None | tuple[str | None, str]]:
        return [
            *await super().get_reference_identifiers(),
            # S3 buckets might reference VPCs through bucket policies or VPC endpoints
            # but we don't have that information in the basic bucket listing
        ]

    async def list_available_metrics(self) -> list[str]:
        return [
            "BucketSizeBytes",
            "NumberOfObjects",
            "AllRequests",
            "GetRequests",
            "PutRequests",
            "DeleteRequests",
            "HeadRequests",
            "PostRequests",
            "SelectRequests",
            "ListRequests",
        ]

    async def get_metric(
        self,
        metric_name: str,
        time_range_start: AwareDatetime,
        time_range_end: AwareDatetime,
    ) -> list[Observation] | str:
        # S3 metrics are available in CloudWatch with different namespaces
        # BucketSizeBytes and NumberOfObjects are in AWS/S3 namespace
        # Request metrics are in AWS/S3 namespace as well

        if metric_name in ["BucketSizeBytes", "NumberOfObjects"]:
            # Storage metrics have different dimensions
            dimensions = [
                {
                    "Name": "BucketName",
                    "Value": self.raw_data["Name"],
                }
            ]
            if metric_name == "BucketSizeBytes":
                dimensions.append({"Name": "StorageType", "Value": "StandardStorage"})
        else:
            # Request metrics
            dimensions = [
                {
                    "Name": "BucketName",
                    "Value": self.raw_data["Name"],
                }
            ]

        return await self._get_cloudwatch_metric_for_accessible_regions(
            Namespace="AWS/S3",
            MetricName=metric_name,
            Dimensions=dimensions,
            StartTime=time_range_start,
            EndTime=time_range_end,
            Period=300,
            Statistics=["Average", "Minimum", "Maximum"]
            if metric_name not in ["BucketSizeBytes", "NumberOfObjects"]
            else ["Average"],
        )



import anyio
import kr8s.asyncio

from daydream.knowledge.graph import Graph
from daydream.plugins import Plugin
from daydream.plugins.kubernetes.nodes.kubernetes_cron_job import KubernetesCronJob
from daydream.plugins.kubernetes.nodes.kubernetes_deployment import KubernetesDeployment
from daydream.plugins.kubernetes.nodes.kubernetes_job import KubernetesJob
from daydream.plugins.kubernetes.nodes.kubernetes_namespace import KubernetesNamespace
from daydream.plugins.kubernetes.nodes.kubernetes_node import KubernetesNode
from daydream.plugins.kubernetes.nodes.kubernetes_pod import KubernetesPod
from daydream.plugins.kubernetes.nodes.kubernetes_replica_set import KubernetesReplicaSet
from daydream.plugins.kubernetes.nodes.kubernetes_service import KubernetesService
from daydream.plugins.kubernetes.nodes.kubernetes_stateful_set import KubernetesStatefulSet
from daydream.plugins.mixins.graph import KnowledgeGraphMixin


class KubernetesPlugin(Plugin, KnowledgeGraphMixin):
    async def validate_plugin_config(self) -> None:
        await super().validate_plugin_config()
        await kr8s.asyncio.version()

    async def populate_graph(self, graph: Graph) -> None:
        await kr8s.asyncio.version()
        print("Populating Kubernetes graph")
        async with anyio.create_task_group() as tg:
            tg.start_soon(self._populate_namespaces, graph)
            tg.start_soon(self._populate_pods, graph)
            tg.start_soon(self._populate_services, graph)
            tg.start_soon(self._populate_deployments, graph)
            tg.start_soon(self._populate_replicasets, graph)
            tg.start_soon(self._populate_statefulsets, graph)
            tg.start_soon(self._populate_jobs, graph)
            tg.start_soon(self._populate_cronjobs, graph)
            tg.start_soon(self._populate_nodes, graph)

    async def _populate_namespaces(self, graph: Graph) -> None:
        async for namespace in kr8s.asyncio.get("namespaces"):
            await graph.add_node(
                KubernetesNamespace(
                    node_id=namespace.metadata.name,
                    raw_data=namespace.to_dict(),
                    _graph=graph,
                )
            )

    async def _populate_pods(self, graph: Graph) -> None:
        async for pod in kr8s.asyncio.get("pods"):
            await graph.add_node(
                KubernetesPod(
                    node_id=pod.metadata.name,
                    raw_data=pod.to_dict(),
                    _graph=graph,
                )
            )

    async def _populate_services(self, graph: Graph) -> None:
        async for service in kr8s.asyncio.get("services"):
            await graph.add_node(
                KubernetesService(
                    node_id=service.metadata.name,
                    raw_data=service.to_dict(),
                    _graph=graph,
                )
            )

    async def _populate_deployments(self, graph: Graph) -> None:
        async for deployment in kr8s.asyncio.get("deployments"):
            await graph.add_node(
                KubernetesDeployment(
                    node_id=deployment.metadata.name,
                    raw_data=deployment.to_dict(),
                    _graph=graph,
                )
            )

    async def _populate_replicasets(self, graph: Graph) -> None:
        async for replicaset in kr8s.asyncio.get("replicasets"):
            await graph.add_node(
                KubernetesReplicaSet(
                    node_id=replicaset.metadata.name,
                    raw_data=replicaset.to_dict(),
                    _graph=graph,
                )
            )

    async def _populate_statefulsets(self, graph: Graph) -> None:
        async for statefulset in kr8s.asyncio.get("statefulsets"):
            await graph.add_node(
                KubernetesStatefulSet(
                    node_id=statefulset.metadata.name,
                    raw_data=statefulset.to_dict(),
                    _graph=graph,
                )
            )

    async def _populate_jobs(self, graph: Graph) -> None:
        async for job in kr8s.asyncio.get("jobs"):
            await graph.add_node(
                KubernetesJob(
                    node_id=job.metadata.name,
                    raw_data=job.to_dict(),
                    _graph=graph,
                )
            )

    async def _populate_cronjobs(self, graph: Graph) -> None:
        async for cronjob in kr8s.asyncio.get("cronjobs"):
            await graph.add_node(
                KubernetesCronJob(
                    node_id=cronjob.metadata.name,
                    raw_data=cronjob.to_dict(),
                    _graph=graph,
                )
            )

    async def _populate_nodes(self, graph: Graph) -> None:
        async for pod in kr8s.asyncio.get("nodes"):
            await graph.add_node(
                KubernetesNode(
                    node_id=pod.metadata.name,
                    raw_data=pod.to_dict(),
                    _graph=graph,
                )
            )



import dns.resolver
import requests
from dns.exception import Timeout as DnsTimeout
from pythonping import ping
from scapy.all import sr1
from scapy.layers.inet import ICMP, IP, UDP

from daydream.plugins import Plugin
from daydream.plugins.mixins import McpServerMixin, tool


class NetworkingPlugin(Plugin, McpServerMixin):
    """A plugin for providing common networking tools."""

    @tool()
    def ping(self, host: str) -> str:
        """Ping a host and return the result."""
        return repr(ping(host, count=4, timeout=5))

    @tool()
    def dig(self, hostname: str, record_type: str = "A") -> str:
        """Use the 'dig' command to resolve a hostname and return the result."""
        result = ""
        try:
            resolver = dns.resolver.Resolver()
            answers = resolver.resolve(hostname, record_type)
            for rdata in answers:
                result += str(rdata) + "\n"
        except dns.resolver.NXDOMAIN:
            result = f"Host '{hostname}' not found."
        except dns.resolver.NoAnswer:
            result = f"No record found for '{hostname}' with type '{record_type}'."
        except DnsTimeout:
            result = "Timeout while querying DNS server."

        return result

    @tool()
    def traceroute(self, destination: str, max_hops: int = 30) -> str:
        """Trace the route to a destination and return the result."""
        result = ""
        for ttl in range(1, max_hops + 1):
            packet = IP(dst=destination, ttl=ttl) / UDP(dport=33434)
            reply = sr1(packet, verbose=0, timeout=1)
            if reply is None:
                result += f"{ttl}\t* * * Request timed out.\n"
            elif reply.haslayer(ICMP) and reply[ICMP].type == 11:
                result += f"{ttl}\t{reply.src}\n"
            elif reply.haslayer(ICMP) and reply[ICMP].type == 3:
                result += f"{ttl}\t{reply.src} Destination reached.\n"
                break
            else:
                result += f"{ttl}\t{reply.src} Other reply type.\n"

        return result

    @tool()
    def request_url(self, url: str) -> str:
        """Request a URL with the GET method and return the result."""
        response = requests.get(url, timeout=10)
        response.raise_for_status()
        return response.text



import os
from typing import Any, cast

from pagerduty.rest_api_v2_client import RestApiV2Client

from daydream.plugins.base import Plugin
from daydream.plugins.mixins import McpServerMixin, tool


class PagerDutyPlugin(Plugin, McpServerMixin):
    _client: RestApiV2Client

    def init_plugin(self) -> None:
        super().init_plugin()
        api_key = self._settings.get("api_key") or os.environ.get("PAGERDUTY_API_KEY", "")
        if api_key:
            self._client = RestApiV2Client(api_key)

    @tool()
    async def get_incident_details(self, incident_id: str) -> dict[str, Any]:
        """Get a PagerDuty incident by ID, including all alert details.

        Args:
            incident_id (str): The ID of the PagerDuty incident. Typically a
            string of uppercase letters and numbers. For example "PGR0VU2",
            "PF9KMXH", or "Q2K78SNJ5U1VE1".

        Returns:
            dict: The incident JSON payload, including all alert details.
        """
        incident = cast("dict[str, Any]", self._client.rget(f"/incidents/{incident_id}"))
        incident["alerts"] = await self.get_alert_details_for_incident(incident_id)
        return incident

    @tool()
    async def get_alert_details_for_incident(self, incident_id: str) -> list[dict[str, Any]]:
        """Get the details of the alert(s) for a PagerDuty incident.

        Args:
            incident_id (str): The ID of the PagerDuty incident. Typically a
            string of uppercase letters and numbers. For example "PGR0VU2",
            "PF9KMXH", or "Q2K78SNJ5U1VE1".

        Returns:
            list[dict]: The list of alert details.
        """
        return cast("list[dict[str, Any]]", self._client.rget(f"/incidents/{incident_id}/alerts"))

    @tool()
    async def post_status_update(self, incident_id: str, message: str) -> None:
        """Post a status update to a PagerDuty incident

        Args:
            incident_id (str): The PagerDuty ID of the incident
            message (str): The message to post
        """
        self._client.rpost(
            f"/incidents/{incident_id}/status_updates",
            json={
                "message": message,
            },
        )



import gzip
import re
from datetime import UTC, datetime
from functools import partial
from multiprocessing import Pool, cpu_count
from pathlib import Path

import msgspec
from pydantic import AwareDatetime

from daydream.models import LogLine

LOG_FILENAME_DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S"


def parse_log_line_timestamp(ts: str) -> AwareDatetime:
    ts_parts = ts.split(".")

    if len(ts_parts) == 2:
        base, nanos = ts_parts
        micros = (nanos[:-1] + "000000")[:6]  # Remove Z, pad/truncate to 6 digits
        ts_fixed = f"{base}.{micros}Z"
        dt = datetime.strptime(ts_fixed, "%Y-%m-%dT%H:%M:%S.%fZ")
    else:
        dt = datetime.strptime(ts, "%Y-%m-%dT%H:%M:%SZ")

    # Make the datetime timezone-aware by assuming UTC
    return dt.replace(tzinfo=UTC)


def process_log_files(
    input_log_filenames: list[Path],
    output_log_filename: Path,
    output_http_log_filename: Path,
    start_dt: AwareDatetime,
    end_dt: AwareDatetime,
) -> None:
    """
    Processes a list of log files, in parallel, and writes them to the
    output files.
    """
    process_fn = partial(
        process_single_log_file,
        output_log_filename=output_log_filename,
        output_http_log_filename=output_http_log_filename,
        start_dt=start_dt,
        end_dt=end_dt,
    )

    # Process log files in parallel using all but one CPU core
    num_processes = max(1, cpu_count() - 1)
    with Pool(num_processes) as pool:
        pool.map(process_fn, input_log_filenames)


def process_single_log_file(
    log_filename: Path,
    output_log_filename: Path,
    output_http_log_filename: Path,
    start_dt: AwareDatetime,
    end_dt: AwareDatetime,
) -> None:
    """
    Determines the type of log and parses it accordingly. Writes the parsed
    log lines to the output file.
    """
    if "proxy" in str(log_filename):
        parse_fn = parse_http_log_line
        output_filename = output_http_log_filename
    else:
        parse_fn = parse_log_line
        output_filename = output_log_filename

    with (
        gzip.open(log_filename, "rt") as input_file,
        Path.open(output_filename, "a") as output_file,
    ):
        for line in input_file:
            parsed = parse_fn(line, start_dt, end_dt)
            if parsed:
                output_file.write(msgspec.json.encode(parsed).decode() + "\n")


def parse_log_line(line: str, start_dt: AwareDatetime, end_dt: AwareDatetime) -> LogLine | None:
    """
    Container logs are already in the correct LogLine format, so parse them
    using msgspec and return them if they are within the time range desired.
    """
    try:
        parsed: LogLine = msgspec.json.decode(line, type=LogLine)
        if start_dt <= parsed.time <= end_dt:
            return parsed
    except (msgspec.DecodeError, ValueError):
        return None


def parse_http_log_line(
    line: str, start_dt: AwareDatetime, end_dt: AwareDatetime
) -> LogLine | None:
    """
    HTTP logs are processed further to extract the method and path, while
    ignoring the rest of the log line (things like IP, user agent, etc) which
    can confuse the template miner.
    """
    try:
        parsed: LogLine = msgspec.json.decode(line, type=LogLine)
        if start_dt <= parsed.time <= end_dt:
            quoted_strings = re.findall(r'"([^"]*)"', parsed.log)
            method_path = quoted_strings[0].lower().split(" http")[0]
            parsed.log = method_path
            return parsed
    except (msgspec.DecodeError, ValueError):
        return None



import typer

from daydream.cli.agent._app import agent_app
from daydream.cli.client._app import client_app
from daydream.cli.tools._app import tools_app

app = typer.Typer(context_settings={"help_option_names": ["-h", "--help"]}, no_args_is_help=True)
app.add_typer(tools_app, name="tools")
app.add_typer(client_app, name="client")
app.add_typer(agent_app, name="agent")



from pydantic import AwareDatetime

from daydream.knowledge import HasMetrics
from daydream.models import Observation

from .base import AwsNode


class AwsRdsDatabase(AwsNode, HasMetrics):
    """An RDS database."""

    async def get_identifiers(self) -> list[str | None]:
        return [
            *await super().get_identifiers(),
            self.raw_data["DBInstanceArn"],
            self.raw_data["DBInstanceIdentifier"],
            self.raw_data["Endpoint"]["Address"],
        ]

    async def get_reference_identifiers(
        self,
    ) -> list[str | None | tuple[str | None, str]]:
        return [
            *await super().get_reference_identifiers(),
            self.raw_data.get("VpcId"),
            *(sg["VpcSecurityGroupId"] for sg in self.raw_data["VpcSecurityGroups"]),
            *(
                (subnet["SubnetIdentifier"], "is_in")
                for subnet in self.raw_data["DBSubnetGroup"]["Subnets"]
            ),
            (
                self.raw_data.get("ReadReplicaSourceDBInstanceIdentifier"),
                "is_replica_of",
            ),
        ]

    async def list_available_metrics(self) -> list[str]:
        return [
            "CPUUtilization",
            "DatabaseConnections",
            "DBLoadRelativeToNumVCPUs",
            "FreeStorageSpace",
            "NetworkReceiveThroughput",
            "ReadIOPS",
            "ReadThroughput",
            "WriteIOPS",
            "WriteThroughput",
        ]

    async def get_metric(
        self,
        metric_name: str,
        time_range_start: AwareDatetime,
        time_range_end: AwareDatetime,
    ) -> list[Observation] | str:
        return await self._get_cloudwatch_metric_for_accessible_regions(
            Namespace="AWS/RDS",
            MetricName=metric_name,
            Dimensions=[
                {
                    "Name": "DBInstanceIdentifier",
                    "Value": self.raw_data["DBInstanceIdentifier"],
                }
            ],
            StartTime=time_range_start,
            EndTime=time_range_end,
            Period=300,
            Statistics=["Average", "Minimum", "Maximum"],
        )



from daydream.plugins.aptible.nodes.base import AptibleNode


class AptibleDisk(AptibleNode):
    async def get_reference_identifiers(
        self,
    ) -> list[str | None | tuple[str | None, str]]:
        return [
            *await super().get_reference_identifiers(),
            (self.raw_data["ebs_volume_id"], "represents"),
            (self.raw_data["host"], "mounted_to"),
        ]



from collections.abc import AsyncGenerator
from typing import Literal

import anyio
import boto3.session
from aioboto3 import Session
from pydantic import BaseModel, Field, ValidationError
from pydantic_core import to_jsonable_python

from daydream.config.utils import Config, PluginSettings
from daydream.knowledge import Graph
from daydream.plugins import Plugin
from daydream.plugins.aws.nodes.aws_alb_target_group import AwsAlbTargetGroup
from daydream.plugins.aws.nodes.aws_application_load_balancer import (
    AwsApplicationLoadBalancer,
)
from daydream.plugins.aws.nodes.aws_classic_load_balancer import AwsClassicLoadBalancer
from daydream.plugins.aws.nodes.aws_ebs_volume import AwsEbsVolume
from daydream.plugins.aws.nodes.aws_ec2_instance import AwsEc2Instance
from daydream.plugins.aws.nodes.aws_rds_database import AwsRdsDatabase
from daydream.plugins.aws.nodes.aws_s3_bucket import AwsS3Bucket
from daydream.plugins.aws.nodes.base import DEFAULT_AWS_ACCOUNT_NAME, AwsAccount
from daydream.plugins.aws.utils import (
    ensure_aws_session,
    list_accessible_regions_for_service,
    swallow_boto_client_access_errors,
)
from daydream.plugins.mixins import KnowledgeGraphMixin
from daydream.utils import Choice, confirm, iprint, print, select


class AwsPluginSettings(BaseModel):
    accounts: dict[str, AwsAccount] = Field(
        default_factory=lambda: {DEFAULT_AWS_ACCOUNT_NAME: AwsAccount()}
    )

    @property
    def account(self) -> AwsAccount:
        return next(iter(self.accounts.values()))


class AwsPlugin(Plugin, KnowledgeGraphMixin):
    aws_settings: AwsPluginSettings = Field(default_factory=AwsPluginSettings)

    def init_plugin(self) -> None:
        aws_accounts = self._settings.get("accounts")
        if not aws_accounts:
            self.aws_settings = AwsPluginSettings()
            return
        if not isinstance(aws_accounts, dict):
            raise ValueError("aws accounts must be a dictionary in config.yaml")
        if len(aws_accounts) != 1:
            raise ValueError(
                "More than one AWS account configured in config.yaml; we only support one AWS account at this time. Please let us know if you have more than one aws account."
            )
        for account_name, account_settings in aws_accounts.items():
            try:
                self.aws_settings = AwsPluginSettings(
                    accounts={
                        account_name: AwsAccount(
                            **{"name": account_name, **to_jsonable_python(account_settings)}
                        )
                    }
                )
            except ValidationError as ex:
                raise ValueError(
                    f"Invalid AWS account settings for aws account '{account_name}'. Review your config.yaml. {account_settings=}; error={ex!s}"
                ) from ex

    async def validate_plugin_config(self) -> None:
        await super().validate_plugin_config()
        await ensure_aws_session(self.session)

    async def interactive_configure(self, cfg: Config) -> PluginSettings:
        iprint("The AWS plugin will add resources from AWS to your infra knowledge graph")
        iprint(
            "You can optionally set an AWS profile from your system that Daydream will use to authenticate AWS requests"
        )
        iprint("Specifying no profile will fallback to using the standard AWS credential providers")
        iprint(
            "Read more about the standard credential providers at: https://docs.aws.amazon.com/sdkref/latest/guide/standardized-credentials.html"
        )
        iprint("")
        avail_aws_profiles = boto3.session.Session().available_profiles
        selected_aws_profile: list[str] = []
        if avail_aws_profiles and await confirm(
            "Would you like to select an AWS profile for Daydream to use when querying your AWS infrastructure?",
            default=True,
        ):
            enable_search = len(avail_aws_profiles) > 10
            selected_aws_profile = [
                await select(
                    "Which AWS profile would you like to use with the Daydream MCP Server?",
                    choices=[Choice(p, checked=(i == 0)) for i, p in enumerate(avail_aws_profiles)],
                    use_search_filter=enable_search,
                    use_jk_keys=not enable_search,
                )
            ]
        settings = AwsPluginSettings(
            accounts={
                profile_name: AwsAccount(
                    name=profile_name,
                    profile=profile_name,
                )
                for profile_name in selected_aws_profile
            }
        )
        return settings.model_dump()

    @property
    def session(self) -> Session:
        if not hasattr(self, "_session"):
            self._session = self.aws_settings.account.session
        return self._session

    async def populate_graph(self, graph: Graph) -> None:
        await ensure_aws_session(self.session)
        async with anyio.create_task_group() as tg:
            tg.start_soon(self.populate_rds_databases, graph)
            tg.start_soon(self.populate_ec2_instances, graph)
            tg.start_soon(self.populate_classic_load_balancers, graph)
            tg.start_soon(self.populate_application_load_balancers, graph)
            tg.start_soon(self.populate_alb_target_groups, graph)
            tg.start_soon(self.populate_ebs_volumes, graph)
            tg.start_soon(self.populate_s3_buckets, graph)

    async def populate_rds_databases(self, graph: Graph) -> None:
        async with anyio.create_task_group() as tg:
            for region in await list_accessible_regions_for_service(self.session, "rds"):
                tg.start_soon(self._populate_rds_databases_in_region, graph, region)

    async def _populate_rds_databases_in_region(self, graph: Graph, region: str) -> None:
        print(f"Populating RDS databases from {region}")

        rds_database_count = 0
        async for page in self._paginate("rds", "describe_db_instances", region):
            for instance in page["DBInstances"]:
                await graph.add_node(
                    AwsRdsDatabase(
                        node_id=instance["DBInstanceArn"],
                        raw_data=instance,
                        _graph=graph,
                        aws_account=self.aws_settings.account,
                    )
                )
                rds_database_count += 1

        print(f"Initialized {rds_database_count} RDS databases for {region}")

    async def populate_ec2_instances(self, graph: Graph) -> None:
        async with anyio.create_task_group() as tg:
            for region in await list_accessible_regions_for_service(self.session, "ec2"):
                tg.start_soon(self._populate_ec2_instances_in_region, graph, region)

    async def _populate_ec2_instances_in_region(self, graph: Graph, region: str) -> None:
        print(f"Populating EC2 instances from {region}")

        ec2_instance_count = 0
        async for page in self._paginate("ec2", "describe_instances", region):
            for reservation in page["Reservations"]:
                for instance in reservation["Instances"]:
                    await graph.add_node(
                        AwsEc2Instance(
                            node_id=instance["InstanceId"],
                            raw_data=instance,
                            _graph=graph,
                            aws_account=self.aws_settings.account,
                        )
                    )
                    ec2_instance_count += 1

        print(f"Initialized {ec2_instance_count} EC2 instances for {region}")

    async def populate_classic_load_balancers(self, graph: Graph) -> None:
        async with anyio.create_task_group() as tg:
            for region in await list_accessible_regions_for_service(self.session, "elb"):
                tg.start_soon(self._populate_classic_load_balancers_in_region, graph, region)

    async def _populate_classic_load_balancers_in_region(self, graph: Graph, region: str) -> None:
        print(f"Populating classic load balancers from {region}")

        elb_count = 0
        async for page in self._paginate("elb", "describe_load_balancers", region):
            for balancer in page["LoadBalancerDescriptions"]:
                await graph.add_node(
                    AwsClassicLoadBalancer(
                        node_id=balancer["LoadBalancerName"],
                        raw_data=balancer,
                        _graph=graph,
                        aws_account=self.aws_settings.account,
                    )
                )
                elb_count += 1

        print(f"Initialized {elb_count} classic load balancers for {region}")

    async def populate_application_load_balancers(self, graph: Graph) -> None:
        async with anyio.create_task_group() as tg:
            for region in await list_accessible_regions_for_service(self.session, "elbv2"):
                tg.start_soon(self._populate_application_load_balancers_in_region, graph, region)

    async def _populate_application_load_balancers_in_region(
        self, graph: Graph, region: str
    ) -> None:
        print(f"Populating application load balancers from {region}")

        alb_count = 0
        async for page in self._paginate("elbv2", "describe_load_balancers", region):
            for balancer in page["LoadBalancers"]:
                await graph.add_node(
                    AwsApplicationLoadBalancer(
                        node_id=balancer["LoadBalancerArn"],
                        raw_data=balancer,
                        _graph=graph,
                        aws_account=self.aws_settings.account,
                    )
                )
                alb_count += 1

        print(f"Initialized {alb_count} application load balancers for {region}")

    async def populate_alb_target_groups(self, graph: Graph) -> None:
        async with anyio.create_task_group() as tg:
            for region in await list_accessible_regions_for_service(self.session, "elbv2"):
                tg.start_soon(self._populate_alb_target_groups_in_region, graph, region)

    async def _populate_alb_target_groups_in_region(self, graph: Graph, region: str) -> None:
        print(f"Populating ALB target groups from {region}")

        alb_target_group_count = 0
        async for page in self._paginate("elbv2", "describe_target_groups", region):
            for target_group in page["TargetGroups"]:
                await graph.add_node(
                    AwsAlbTargetGroup(
                        node_id=target_group["TargetGroupArn"],
                        raw_data=target_group,
                        _graph=graph,
                        aws_account=self.aws_settings.account,
                    )
                )
                alb_target_group_count += 1

        print(f"Initialized {alb_target_group_count} ALB target groups for {region}")

    async def populate_ebs_volumes(self, graph: Graph) -> None:
        async with anyio.create_task_group() as tg:
            for region in await list_accessible_regions_for_service(self.session, "ec2"):
                tg.start_soon(self._populate_ebs_volumes_in_region, graph, region)

    async def _populate_ebs_volumes_in_region(self, graph: Graph, region: str) -> None:
        print(f"Populating EBS volumes from {region}")

        ebs_volume_count = 0
        async for page in self._paginate("ec2", "describe_volumes", region):
            for volume in page["Volumes"]:
                await graph.add_node(
                    AwsEbsVolume(
                        node_id=volume["VolumeId"],
                        raw_data=volume,
                        _graph=graph,
                        aws_account=self.aws_settings.account,
                    )
                )
                ebs_volume_count += 1

        print(f"Initialized {ebs_volume_count} EBS volumes for {region}")

    async def populate_s3_buckets(self, graph: Graph) -> None:
        # S3 buckets are global resources, not regional like other AWS services
        # We only need to call list_buckets once, not per region
        print("Populating S3 buckets")

        s3_bucket_count = 0
        async with (
            swallow_boto_client_access_errors(service_name="s3", region="us-east-1"),
            self.session.client("s3", region_name="us-east-1") as client,
        ):
            paginator = client.get_paginator("list_buckets")
            async for page in paginator.paginate():
                for bucket in page["Buckets"]:
                    await graph.add_node(
                        AwsS3Bucket(
                            node_id=bucket["Name"],
                            raw_data=bucket,
                            _graph=graph,
                            aws_account=self.aws_settings.account,
                        )
                    )
                    s3_bucket_count += 1

        print(f"Initialized {s3_bucket_count} S3 buckets")

    async def _paginate(
        self,
        service_name: Literal["ec2", "elb", "elbv2", "rds"],
        action: str,
        region: str,
    ) -> AsyncGenerator[dict, None]:
        async with (
            swallow_boto_client_access_errors(service_name=service_name, region=region),
            self.session.client(service_name, region_name=region) as client,
        ):
            paginator = client.get_paginator(action)
            async for page in paginator.paginate():
                yield page



import os
from datetime import timedelta
from pathlib import Path
from typing import TYPE_CHECKING

from pydantic import AwareDatetime, BaseModel
from sh import Command, ErrorReturnCode

from daydream.utils import print

if TYPE_CHECKING:
    from daydream.plugins.aptible.nodes.aptible_container import AptibleContainer

S3_LOGS_DATE_FORMAT = "%Y-%m-%d"
SWEETNESS_LOG_BUCKET_NAME_SETTING = "sweetness/nonsensitive/DOCKER_LOGS_BUCKET_NAME"
SWEETNESS_LOG_ENCRYPTION_KEYS_SETTING = "sweetness/sensitive/LOG_ENCRYPTION_KEYS"
NO_FILES_FOUND_ERROR = "No files found that matched all criteria"


class LogBucketConfig(BaseModel):
    bucket_name: str
    region: str
    decryption_keys: str


async def get_log_bucket_config(stack: str) -> LogBucketConfig:
    log_bucket = await get_pancake_setting(stack, SWEETNESS_LOG_BUCKET_NAME_SETTING)
    log_encryption_keys = await get_pancake_setting(stack, SWEETNESS_LOG_ENCRYPTION_KEYS_SETTING)

    return LogBucketConfig(
        bucket_name=log_bucket,
        region=log_bucket.replace("aptible-docker-logs-", ""),
        decryption_keys=log_encryption_keys,
    )


async def download_s3_logs(
    stack: str,
    container: "AptibleContainer",
    start_dt: AwareDatetime,
    end_dt: AwareDatetime,
    download_dir: Path,
    bucket_config: LogBucketConfig,
) -> int:
    print(
        f"Downloading s3 logs for {container.raw_data['docker_name']} ({container.raw_data['layer']})"
    )

    # S3 logs can only be filtered by day, not time - so add 1 day to end_dt
    # to make sure we get all logs for the last day
    start_date = start_dt.date()
    end_date = (end_dt + timedelta(days=1)).date()

    try:
        aptible_cli = Command("aptible")
        output = await aptible_cli.logs_from_archive(
            stack=stack,
            container_id=container.raw_data["docker_name"],
            start_date=start_date.strftime(S3_LOGS_DATE_FORMAT),
            end_date=end_date.strftime(S3_LOGS_DATE_FORMAT),
            region=bucket_config.region,
            bucket=bucket_config.bucket_name,
            decryption_keys=bucket_config.decryption_keys,
            download_location=str(download_dir),
            _async=True,
            _err_to_out=True,
        )

        files_downloaded = 0
        output_lines = output.splitlines()
        for line in output_lines:
            if download_dir.as_posix() in line:
                files_downloaded += 1

        return files_downloaded
    except ErrorReturnCode as e:
        if NO_FILES_FOUND_ERROR not in str(e):
            print(f"Error running command: {e}")
        return 0


async def get_pancake_setting(stack: str, setting: str) -> str:
    pancake = Command(os.getenv("PANCAKE_PATH"))

    try:
        result = await pancake(  # pyright: ignore [reportGeneralTypeIssues]
            "stack:settings:view_single",
            stack,
            setting,
            _async=True,
        )
        return result.strip()
    except ErrorReturnCode as e:
        print(f"Error running command: {e}")

    return ""



import asyncio
import random
import re
import shutil
import warnings
from collections.abc import Callable, Coroutine
from datetime import UTC, datetime
from pathlib import Path
from typing import TYPE_CHECKING

import msgspec
from pydantic import AwareDatetime

from daydream.models import LogLine
from daydream.plugins.aptible.logs.processing import LOG_FILENAME_DATETIME_FORMAT
from daydream.plugins.aptible.nodes.base import AptibleNode

from .container import download_container_logs
from .processing import process_log_files
from .s3 import download_s3_logs, get_log_bucket_config

if TYPE_CHECKING:
    from daydream.plugins.aptible.nodes.aptible_container import AptibleContainer


LOGS_DIR = Path.home() / ".daydream" / "logs"
LOG_FILENAME_PATTERN = re.compile(r"^.*\.archived\.gz$")
PROCESSED_LOGS_FILENAME = "logs.jsonl"
PROCESSED_HTTP_LOGS_FILENAME = "http-logs.jsonl"
MAX_CONTAINERS_BEFORE_SAMPLING = 20
INITIAL_DOWNLOAD_ATTEMPTS = 2
MAX_CONCURRENT_CONTAINER_DOWNLOADS = 3
MAX_CONCURRENT_S3_DOWNLOADS = 20
MAX_EMPTY_DOWNLOADS = 2


class DownloadMonitor:
    downloads: int
    none_found: int

    def __init__(self) -> None:
        self.downloads = 0
        self.none_found = 0


async def download_logs_for_time_range(
    stack: str,
    node: AptibleNode,
    containers: list["AptibleContainer"],
    start_dt: AwareDatetime,
    end_dt: AwareDatetime,
) -> list[LogLine]:
    """
    Download logs for a given time range.

    Logs are downloaded from S3 and from live containers in parallel. Once
    downloaded, they are processed and cached.

    Returns a list of LogLine objects.
    """
    # Create a directory to store the logs
    log_dir = _log_directory_name(stack, node, start_dt, end_dt)
    log_dir.mkdir(parents=True, exist_ok=True)
    log_filename = log_dir / PROCESSED_LOGS_FILENAME
    http_log_filename = log_dir / PROCESSED_HTTP_LOGS_FILENAME

    # Check for cached logs first, which are pre-processed and ready to go.
    # Otherwise, download the logs and process them.
    if not _log_files_downloaded(log_dir):
        tmp_download_dir = await download_s3_and_container_logs(
            stack, containers, start_dt, end_dt, log_dir
        )

        relevant_log_files = _log_files_for_time_range(tmp_download_dir, start_dt, end_dt)
        process_log_files(relevant_log_files, log_filename, http_log_filename, start_dt, end_dt)
        shutil.rmtree(tmp_download_dir)

    # Read processed container & HTTP logs into a list of LogLine objects
    log_lines: list[LogLine] = []

    if log_filename.exists():
        with Path.open(log_filename, "r") as f:
            log_lines.extend(msgspec.json.decode(line, type=LogLine) for line in f)

    if http_log_filename.exists():
        with Path.open(http_log_filename, "r") as f:
            log_lines.extend(msgspec.json.decode(line, type=LogLine) for line in f)

    return log_lines


async def download_s3_and_container_logs(
    stack: str,
    containers: list["AptibleContainer"],
    start_dt: AwareDatetime,
    end_dt: AwareDatetime,
    log_dir: Path,
) -> Path:
    """
    Downloads logs from S3 and live containers in parallel into a temporary
    directory. Logs are split into app and proxy (vhost) logs.

    Downloads for a specific (source, container_layer) pair are monitored. If
    there are no logs from a source after MAX_EMPTY_DOWNLOADS, that means the
    logs are probably in another source, so we can stop early.

    The number of parallel downloads are limited by source, ie containers and
    s3 have their own semaphores.

    Returns the path to the temporary directory where logs were downloaded.
    """
    app_containers, proxy_containers = _prepare_containers(containers)
    bucket_config = await get_log_bucket_config(stack)

    # Download unprocessed logs to a temporary directory
    tmp_download_dir = log_dir / "tmp"
    tmp_download_dir.mkdir(parents=True, exist_ok=True)

    # Monitor downloads. If there are no logs from a source after
    # MAX_EMPTY_DOWNLOADS, that means the logs are probably in
    # another source, and we can stop early.
    monitor_app_container_downloads = generate_monitor_fn(DownloadMonitor())
    monitor_proxy_container_downloads = generate_monitor_fn(DownloadMonitor())
    monitor_app_s3_downloads = generate_monitor_fn(DownloadMonitor())
    monitor_proxy_s3_downloads = generate_monitor_fn(DownloadMonitor())

    # Create a semaphore to limit concurrent downloads by source
    container_semaphore = asyncio.Semaphore(MAX_CONCURRENT_CONTAINER_DOWNLOADS)
    s3_semaphore = asyncio.Semaphore(MAX_CONCURRENT_S3_DOWNLOADS)

    # Download all logs in parallel. Concurrency is limited by source, ie
    # containers and s3 have their own semaphores. For each (source, container layer)
    # pair, we monitor for empty downloads and stop when we get > MAX_EMPTY_DOWNLOADS.
    # We do this because sometimes an app's logs are in one source, while its
    # proxy logs are in another, even for the same time range.
    await asyncio.gather(
        *(
            monitor_app_container_downloads(
                download_container_logs(stack, container, start_dt, end_dt, tmp_download_dir),
                container_semaphore,
            )
            for container in app_containers
        ),
        *(
            monitor_proxy_container_downloads(
                download_container_logs(stack, container, start_dt, end_dt, tmp_download_dir),
                container_semaphore,
            )
            for container in proxy_containers
        ),
        *(
            monitor_app_s3_downloads(
                download_s3_logs(
                    stack, container, start_dt, end_dt, tmp_download_dir, bucket_config
                ),
                s3_semaphore,
            )
            for container in app_containers
        ),
        *(
            monitor_proxy_s3_downloads(
                download_s3_logs(
                    stack, container, start_dt, end_dt, tmp_download_dir, bucket_config
                ),
                s3_semaphore,
            )
            for container in proxy_containers
        ),
    )

    return tmp_download_dir


def _prepare_containers(
    containers: list["AptibleContainer"],
) -> tuple[list["AptibleContainer"], list["AptibleContainer"]]:
    """
    Split containers into app and proxy containers. If there are too many
    containers, sample them.
    """
    if len(containers) > MAX_CONTAINERS_BEFORE_SAMPLING:
        containers = random.sample(containers, MAX_CONTAINERS_BEFORE_SAMPLING)

    app_containers = [c for c in containers if c.raw_data["layer"] == "app"]
    proxy_containers = [c for c in containers if c.raw_data["layer"] == "proxy"]
    return app_containers, proxy_containers


def generate_monitor_fn(
    monitor: DownloadMonitor,
) -> Callable[[Coroutine, asyncio.Semaphore], Coroutine[None, None, None]]:
    """
    Generate a function that monitors downloads for a given source, watching
    for empty downloads and limiting concurrent downloads with a semaphore.
    """

    async def monitor_downloads(coro: Coroutine, semaphore: asyncio.Semaphore) -> None:
        # Ignore warnings about the coroutine not being awaited when exiting early
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            async with semaphore:
                if (
                    monitor.downloads >= INITIAL_DOWNLOAD_ATTEMPTS
                    and monitor.none_found >= MAX_EMPTY_DOWNLOADS
                ):
                    return

                files_downloaded = await coro
                monitor.downloads += 1
                if files_downloaded == 0:
                    monitor.none_found += 1

    return monitor_downloads


def _log_directory_name(
    stack: str,
    node: AptibleNode,
    start_dt: AwareDatetime,
    end_dt: AwareDatetime,
) -> Path:
    return LOGS_DIR / "-".join(
        [
            stack,
            node.raw_data["_type"],
            node.node_id,
            start_dt.strftime(LOG_FILENAME_DATETIME_FORMAT),
            end_dt.strftime(LOG_FILENAME_DATETIME_FORMAT),
        ]
    )


def _log_files_downloaded(log_dir: Path) -> bool:
    return (log_dir / PROCESSED_LOGS_FILENAME).exists() or (
        log_dir / PROCESSED_HTTP_LOGS_FILENAME
    ).exists()


def _log_files_for_time_range(
    log_dir: Path, start_dt: AwareDatetime, end_dt: AwareDatetime
) -> list[Path]:
    """
    Downloads from S3 can only be limited by YYY-MM-DD, and so we have to download
    files for time_range + 1 day to make sure we have full coverage. To avoid
    unnecessary log file parsing, this function uses the time range from the
    filename to check for overlap first.
    """
    filenames = [
        f for f in log_dir.glob("**/*") if f.is_file() and LOG_FILENAME_PATTERN.match(f.name)
    ]
    matching_files = []
    for filename in filenames:
        dt_range = str(filename).split("json.log.")[1].replace(".archived.gz", "")
        dt_range = dt_range.removeprefix("1.")

        log_start, log_end = dt_range.split(".")
        log_start_dt = datetime.strptime(log_start, LOG_FILENAME_DATETIME_FORMAT).replace(
            tzinfo=UTC
        )
        log_end_dt = datetime.strptime(log_end, LOG_FILENAME_DATETIME_FORMAT).replace(tzinfo=UTC)

        # Check if the log file's time range overlaps with our target period
        if (
            (start_dt <= log_start_dt <= end_dt)
            or (start_dt <= log_end_dt <= end_dt)
            or (log_start_dt <= start_dt <= log_end_dt)
        ):
            matching_files.append(filename)

    return matching_files



import anyio
from rich import print

from daydream.agent.utils import get_agent_templates
from daydream.cli.agent._app import agent_app


@agent_app.command()
def templates() -> None:
    """List the available agent templates."""

    async def _run() -> None:
        print("Available agent templates:")
        for template in sorted(get_agent_templates()):
            print(f"* {template}")

    anyio.run(_run)



from .base import KubernetesBaseNode


class KubernetesNamespace(KubernetesBaseNode):
    async def get_identifiers(self) -> list[str | None]:
        return [
            *await super().get_identifiers(),
        ]

    async def get_reference_identifiers(self) -> list[str | None | tuple[str | None, str]]:
        return [
            *await super().get_reference_identifiers(),
        ]



import os
from typing import Any

import anyio
import questionary

from daydream.config.utils import Config, PluginSettings
from daydream.knowledge import Graph
from daydream.plugins.base import Plugin
from daydream.plugins.datadog.client import DatadogClient
from daydream.plugins.datadog.nodes.datadog_api import DatadogApi
from daydream.plugins.datadog.nodes.datadog_service import DatadogService
from daydream.plugins.datadog.nodes.datadog_system import DatadogSystem
from daydream.plugins.datadog.nodes.datadog_team import DatadogTeam
from daydream.plugins.mixins import KnowledgeGraphMixin


class DatadogPlugin(Plugin, KnowledgeGraphMixin):
    """A plugin for Datadog."""

    _client: DatadogClient

    def __init__(
        self,
        *args: Any,
        api_key: str | None = None,
        application_key: str | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._api_key = api_key
        self._application_key = application_key
        self._client = DatadogClient(
            api_key=api_key or os.getenv("DATADOG_API_KEY", ""),
            application_key=application_key or os.getenv("DATADOG_APP_KEY", ""),
        )

    async def interactive_configure(self, cfg: Config) -> PluginSettings:
        """Interactive wizard for configuring the settings of this plugin."""
        return {
            "api_key": await questionary.text("Enter your Datadog API key").unsafe_ask_async(),
            "application_key": await questionary.text(
                "Enter your Datadog application key"
            ).unsafe_ask_async(),
        }

    async def populate_graph(self, graph: Graph) -> None:
        """Populate the graph with Datadog entities."""
        async with anyio.create_task_group() as tg:
            tg.start_soon(self._populate_teams, graph)
            tg.start_soon(self._populate_software_catalog_entities, graph)

    async def _populate_teams(self, graph: Graph) -> None:
        async for team in self._client.iter_teams():
            await graph.add_node(
                DatadogTeam(
                    node_id=f"team:default/{team['attributes']['handle']}",
                    raw_data=team.to_dict(),
                    _graph=graph,
                )
            )

    async def _populate_software_catalog_entities(self, graph: Graph) -> None:
        async for entity in self._client.iter_software_catalog_entities():
            entity_kind = entity["attributes"]["kind"]

            if entity_kind == "service":
                node = DatadogService(
                    node_id=f"service:default/{entity['attributes']['name']}",
                    raw_data=entity.to_dict(),
                    _graph=graph,
                )
            elif entity_kind == "api":
                node = DatadogApi(
                    node_id=f"api:default/{entity['attributes']['name']}",
                    raw_data=entity.to_dict(),
                    _graph=graph,
                )
            else:
                raise ValueError(f"Unknown entity kind: {entity_kind}")

            await graph.add_node(node)

            # HACK: There doesn't seem to be a way to list systems directly in datadog, so we'll fake it for now.
            for (
                related_entity_reference,
                _,
            ) in await node._get_related_entity_reference_identifiers():
                if related_entity_reference.startswith("system:"):
                    await graph.add_node(
                        DatadogSystem(
                            node_id=related_entity_reference,
                            raw_data={
                                "attributes": {
                                    "name": related_entity_reference.split("/")[-1],
                                },
                            },
                            _graph=graph,
                        )
                    )



from daydream.plugins.aws.nodes.base import AwsNode


class AwsEbsVolume(AwsNode):
    async def get_identifiers(self) -> list[str | None]:
        return [
            *await super().get_identifiers(),
            self.raw_data["VolumeId"],
        ]



from datetime import timedelta
from decimal import Decimal
from typing import Literal

from pydantic import AwareDatetime, BaseModel, Field, computed_field, field_validator


class Point(BaseModel):
    """A data point in a series."""

    timestamp: AwareDatetime
    value: Decimal


class Anomaly(BaseModel):
    points: list[Point] = Field(description="The points that make up the anomaly")
    baseline: float = Field(
        description="The baseline value for the anomaly (i.e., what the anomaly was compared against)"
    )
    significance_scores: list[float] = Field(
        description="A measure of the significance of the anomaly (higher is more significant)"
    )
    description: Literal["high", "low", "jittery"] = Field(description="The nature of the anomaly")

    @computed_field  # type: ignore[prop-decorator]
    @property
    def max_significance(self) -> float:
        return max(self.significance_scores)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def min_significance(self) -> float:
        return min(self.significance_scores)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def total_significance(self) -> float:
        return sum(self.significance_scores)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def average_significance(self) -> float:
        return self.total_significance / len(self.points)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def start_time(self) -> AwareDatetime:
        return self.points[0].timestamp

    @computed_field  # type: ignore[prop-decorator]
    @property
    def end_time(self) -> AwareDatetime:
        return self.points[-1].timestamp

    @computed_field  # type: ignore[prop-decorator]
    @property
    def min_value(self) -> Decimal:
        return min(p.value for p in self.points)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def max_value(self) -> Decimal:
        return max(p.value for p in self.points)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def duration(self) -> timedelta:
        return self.end_time - self.start_time

    @field_validator("points", mode="before")
    @classmethod
    def validate_points(cls, v: list[Point]) -> list[Point]:
        """Validate that the points are sorted by timestamp."""
        return sorted(v, key=lambda x: x.timestamp)

    def __add__(self, other: "Anomaly") -> "Anomaly":
        """Merge two anomalies."""
        return Anomaly(
            points=self.points + other.points,
            significance_scores=self.significance_scores + other.significance_scores,
            baseline=(self.baseline + other.baseline) / 2,
            description=self.description if self.description == other.description else "jittery",
        )


def detect_anomalies(data: list[Point], contamination: float = 0.01) -> list[Anomaly]:
    """
    Detect anomalies in a list of (timestamp, value) tuples using pyod's Isolation Forest model.
    Also calculates a significance score for each anomaly based on its deviation from normal.

    :param data: A list of (timestamp, value) tuples.
    :param contamination: The proportion of outliers in the data set (default: 0.01).
    :return: A list of Anomaly objects that are flagged as anomalies.
    """
    import numpy as np
    from pyod.models.iforest import IForest

    if len(data) < 5:
        return []

    # Convert points to numpy array with float dtype
    points = np.array([[float(p.value)] for p in data], dtype=np.float64)

    # Handle constant values or invalid data
    if np.all(points == points[0]) or not np.all(np.isfinite(points)):
        return []

    # Calculate the range of values and the mean
    value_range = np.max(points) - np.min(points)
    mean_value = np.mean(points)

    # Values this small are effectively constant for our purposes.
    if value_range < 1e-10:
        return []

    # Use pyod's Isolation Forest to detect anomalies.
    clf = IForest(contamination=contamination, random_state=42)
    clf.fit(points)

    # Get anomaly scores (approximately -0.5 to 0.5 for IForest) and predictions (1 for outliers, 0 for inliers)
    anomaly_scores = clf.decision_function(points)
    predictions = clf.predict(points)

    if not anomaly_scores:
        return []

    # Normalize scores to 0-1 range if there are anomalies
    if np.ptp(anomaly_scores) > 0:
        anomaly_scores = (anomaly_scores - anomaly_scores.min()) / (
            anomaly_scores.max() - anomaly_scores.min()
        )

    # Return anomalous points with their significance scores and direction
    return [
        Anomaly(
            points=[point],
            significance_scores=[score],
            baseline=float(mean_value),
            description="high" if float(point.value) > mean_value else "low",
        )
        for point, score, is_anomaly in zip(data, anomaly_scores, predictions, strict=True)
        if is_anomaly
    ]


def cluster_anomalies(anomalies: list[Anomaly], threshold_seconds: int = 1800) -> list[Anomaly]:
    """
    Group anomalies into clusters based on a time threshold.

    :param anomalies: A list of Anomaly objects.
    :param threshold_seconds: Maximum gap in seconds to keep anomalies in the same cluster.
    :return: A list of Anomaly objects, where each Anomaly contains multiple points.
    """
    clusters: list[Anomaly] = []

    if not anomalies:
        return clusters

    current_cluster: Anomaly | None = None
    for anomaly in sorted(anomalies, key=lambda x: x.start_time):
        if current_cluster is None:
            # Start a new cluster with the first anomaly
            current_cluster = anomaly
        else:
            time_diff = (anomaly.start_time - current_cluster.end_time).total_seconds()

            # If the anomaly is within the threshold, merge it with the current cluster.
            if time_diff <= threshold_seconds:
                current_cluster += anomaly
            else:
                # Otherwise, finalize the current cluster and start a new one
                clusters.append(current_cluster)
                current_cluster = anomaly

    # Add the final cluster if it exists
    if current_cluster is not None:
        clusters.append(current_cluster)

    return clusters



import anyio
import typer

from daydream.cli._app import app
from daydream.cli.options import PROFILE_OPTION
from daydream.config.utils import get_config_dir
from daydream.knowledge import Graph
from daydream.telemetry import client as telemetry


@app.command()
def visualize(
    profile: str = PROFILE_OPTION,
    topology: bool = typer.Option(False, "--topology"),
) -> None:
    """Visualize the knowledge graph topology"""

    async def _visualize() -> None:
        await telemetry.send_event(
            {
                "command": "visualize_graph",
                "profile": profile,
                "topology_only": topology,
            }
        )

        graph = Graph(get_config_dir(profile) / "graph.json")

        if topology:
            graph = await graph.get_topology()

        print(await graph.to_pydot())

    anyio.run(_visualize)



import os
from pathlib import Path

import anyio
import typer
from rich import print

from daydream.agent.utils import get_agent_template
from daydream.cli.agent._app import agent_app
from daydream.cli.options import PROFILE_OPTION
from daydream.config.utils import get_config_dir
from daydream.utils import edit_file


@agent_app.command()
def create(
    agent_name: str = typer.Argument(..., help="The name of the agent to create"),
    profile: str = PROFILE_OPTION,
    overwrite: bool = typer.Option(False, help="Overwrite the agent file if it already exists"),
    template: str = typer.Option("blank", help="The template to use to create the agent file"),
    editor: str = typer.Option(
        os.environ.get("EDITOR"), help="The editor to use to open the agent file"
    ),
    no_edit: bool = typer.Option(False, help="Do not open the agent file in your editor"),
) -> None:
    """Create a new agent configuration file and open it in your editor."""

    async def _create() -> None:
        # Create the default YAML content
        try:
            agent_template = get_agent_template(template)
        except FileNotFoundError as ex:
            print(
                f"Template '{template}' not found at {Path(__file__).parent / 'templates' / f'{template}.yaml'}"
            )
            raise typer.Abort() from ex

        # Get the config directory for the profile
        config_dir = get_config_dir(profile, create=True)

        # Create the agents directory if it doesn't exist
        agents_dir = config_dir / "agents"
        agents_dir.mkdir(exist_ok=True)

        # Create the agent file path
        agent_file = agents_dir / f"{agent_name}.yaml"

        # Check if the agent file already exists
        if agent_file.exists():
            if overwrite:
                print(f"Overwriting agent '{agent_name}' at {agent_file}")
            else:
                print(f"Agent '{agent_name}' already exists at {agent_file}")
                if not typer.confirm("Do you want to overwrite it?"):
                    raise typer.Abort()

        # Write the YAML content to the file
        agent_file.write_text(agent_template, encoding="utf-8")

        print(f"Created agent configuration at {agent_file}")

        # Open the file in the user's editor
        if editor and not no_edit:
            try:
                await edit_file(agent_file, editor)
            except ValueError as ex:
                print(
                    "[red]No editor specified. Set the $EDITOR environment variable or use --editor option.[/red]"
                )
                print(f"[blue]Please manually open {str(agent_file)!r} in your editor.[/blue]")
                raise typer.Exit() from ex

    anyio.run(_create)
